from collections import defaultdict
from logging import getLogger
from queue import PriorityQueue
from time import sleep, time
from threading import Thread, Lock
from typing import Callable


logger = getLogger('task_queue')


class TaskQueue:
  """Class to asynchronously handle tasks in the background."""

  def __init__(self, num_workers: int = 1):
    """Initialize the task queue.

    Args;
      num_workers: How many worker threads to launch to process tasks.
    """
    self.num_workers = num_workers
    self.queue = PriorityQueue()
    # Map from task id to latest version number for that task
    self.tasks = defaultdict(int)
    self.task_locks = defaultdict(Lock)

  def run_worker(self, i, num_retries=5):
    """Function each worker will run.

    Args:
      i: The thread index.
      num_retries; How many times to retry the task.
    """
    logger.info('Initialized task worker %s', i)
    while True:
      task = self.queue.get()
      time_to_run, task_id, task_version, fn, args, kwargs = task
      logger.info('Received task %s', task_id)
      logger.info('Task queue size %s', self.queue.qsize())

      # If there is a newer version of the task, skip this one
      with self.task_locks[task_id]:
        if self.tasks[task_id] > task_version:
          logger.info('Task cancelled')
          self.queue.task_done()
          continue

      # Sleep until we are ready to run the code
      time_to_sleep = min(1, time_to_run - time())
      logger.info('Time to sleep %s', time_to_sleep)
      if time_to_sleep > 0:
        sleep(time_to_sleep)

      # Make this check again
      with self.task_locks[task_id]:
        if self.tasks[task_id] > task_version:
          logger.info('Task cancelled')
          self.queue.task_done()
          continue

      # Run the function, retry on failures
      finished = False
      for backoff in [2**i for i in range(num_retries + 1)]:
        try:
          fn(*args, **kwargs)
          finished = True
          break
        except:
          logger.info('An error occurred, sleeping %s', backoff)
          sleep(backoff)

      # Put the task back in the queue if we still failed
      if not finished:
        logger.info('Task failed, reinserting into queue %s', task_id)
        self.queue.put(task)

      self.queue.task_done()

  def start(self):
    """Start the background worker threads."""
    for i in range(self.num_workers):
      thread = Thread(target=self.run_worker, args=(i,))
      thread.start()

  def submit_task(self, task_id: str, delay: float, fn: Callable, *args,
                  **kwargs):
    """Add a task to run.

    Args:
      task_id: An id to specify the task.
      delay: How much time to wait before running the task.
      fn: The function to run.
      args: The args to pass to fn.
      kwargs: The kwargs to pass to fn.
    """
    logger.info('Received task %s %s', task_id, delay)
    time_to_run = time() + delay
    args = args or ()
    kwargs = kwargs or {}

    with self.task_locks[task_id]:
      self.tasks[task_id] += 1
      task_version = self.tasks[task_id]

    self.queue.put((time_to_run, task_id, task_version, fn, args, kwargs))
    logger.info('Task queue size %s', self.queue.qsize())

  def cancel_task(self, task_id: str):
    """Cancel a submitted task.

    Args:
      task_id: The task to cancel.
    """
    logger.info('Cancel task %s', task_id)
    with self.task_locks[task_id]:
      self.tasks[task_id] += 1



import os
from glob import glob
from logging import getLogger

from b2py import utils as b2_utils


logger = getLogger('cache')


class Cache:
  """Cache files to the local disk to save bandwidth."""

  def __init__(self, cache_dir: str, cache_size: int):
    """Initialize a local object cache.

    Args:
      cache_dir: The directory to save cached files to.
      cache_size: The size, in MB, to limit the cache dir to.
    """
    logger.info('Initializing cache')
    self.cache_dir = cache_dir
    self.cache_size = int(cache_size * 1e6)
    self.index = {}
    self.touch_list = []
    self._populate_index()

  def _populate_index(self):
    """Read the cache dir and set a local index of records."""
    os.makedirs(self.cache_dir, exist_ok=True)
    local_files = glob('{}/*'.format(self.cache_dir))
    for file in local_files:
      self._add_to_index(os.path.basename(file), os.path.getsize(file))

  def _touch_file(self, file_id):
    """Move the file to the end of the queue by touching it.

    Args:
      file_id: The file to touch.
    """
    if file_id in self.touch_list:
      self.touch_list.remove(file_id)
    self.touch_list.append(file_id)

  def _recover_disk_space(self):
    """Make sure we stay under our disk space quota."""
    while self.used_disk_space > self.cache_size:
      space_to_recover = self.used_disk_space - self.cache_size
      logger.info('Recovering disk space %s', space_to_recover)
      lru_file = self.touch_list.pop(0)
      file_path = self._path_to_file(lru_file)
      logger.info('Deleting %s', file_path)
      os.remove(file_path)
      del self.index[lru_file]

  def _path_to_file(self, file_id: str):
    """
    Args:
      file_id: The B2 file id to query.

    Returns:
      The local path to the cached file.
    """
    return os.path.join(self.cache_dir, file_id)

  @property
  def used_disk_space(self) -> int:
    """
    Returns:
      The used disk space in bytes.
    """
    return sum(self.index.values())

  def _add_to_index(self, file_id: str, content_size: int):
    """
    Args:
      file_id: The file key.
      content_size: The size of the file's contents.
    """
    self._touch_file(file_id)
    self.index[file_id] = content_size
    self._recover_disk_space()

  def has(self, file_id):
    """
    Args:
      file_id: The file to check.

    Returns:
      Whether the cache contains the file.
    """
    return file_id in self.index

  def add(self, file_id: str, contents: bytes):
    """Add a file to the cache.

    Args:
      file_id: The unique key to look the file up by.
      contents: The file contents.
    """
    file_path = self._path_to_file(file_id)
    b2_utils.write_file(file_path, contents)
    self._add_to_index(file_id, len(contents))

  def get(self, file_id: str) -> bytes:
    """
    Args:
      file_id: The file to read
    
    Returns:
      The file's contents.
    """
    if not self.has(file_id):
      raise KeyError('No local copy of file {}'.format(file_id))
    self._touch_file(file_id)
    file_path = self._path_to_file(file_id)
    return b2_utils.read_file(file_path)

  def delete(self, file_id: str):
    """Delete the file from the cache.

    Args:
      file_id: The file to delete.
    """
    file_path = self._path_to_file(file_id)
    os.remove(file_path)
    del self.index[file_id]



from collections import defaultdict
from errno import ENOENT, ENOTEMPTY, EINVAL
from logging import getLogger
from stat import S_IFDIR, S_IFLNK
from threading import Lock
from time import time
from typing import Dict, List, Tuple, Union

from b2py import B2, utils as b2_utils
from fuse import FuseOSError, Operations, LoggingMixIn

from zerofs.cache import Cache
from zerofs.file import Directory, File
from zerofs.task_queue import TaskQueue


logger = getLogger('zerofs')


class ZeroFS(LoggingMixIn, Operations):
  """Virtual filesystem backed by the B2 object store."""

  def __init__(self, bucket_name: str, cache_dir: str, cache_size: int,
               upload_delay: float, num_workers: int):
    """Initialize the FUSE filesystem.

    Args:
      bucket_name: The name of the remote bucket to mount.
      cache_dir: The directory to cache files to.
      cache_size: The cache size in MB for saving files on local disk.
      upload_delay: Delay in seconds after writing before uploading to cloud.
      num_workers: Number of background thread workers.
    """
    logger.info('Initializing zerofs from bucket {}'.format(bucket_name))
    self.bucket_name = bucket_name
    self.cache = Cache(cache_dir, cache_size)
    self.b2 = B2()
    self.file_locks = defaultdict(Lock)
    self.upload_delay = upload_delay

    # Load the directory tree
    logger.info('Loading directory tree')
    self._load_dir_tree()

    # Start the task queue
    logger.info('Starting task queue')
    self.task_queue = TaskQueue(num_workers)
    self.task_queue.start()

  @staticmethod
  def _to_bytes(s: Union[str, bytes]):
    if type(s) == bytes:
      return s
    return s.encode('utf-8')

  def _load_dir_tree(self):
    """Load the directory structure into memory."""
    buckets = self.b2.list_buckets()
    bucket = [b for b in buckets if b['bucketName'] == self.bucket_name]
    if not len(bucket):
      raise ValueError('Create a bucket named {} to enable zerofs.'.format(
          self.bucket_name))
    self.bucket_id = bucket[0]['bucketId']
    files = [File(f) for f in self.b2.list_files(self.bucket_id, limit=10000)]
    self.root = Directory('', files)
    self.fd = 0

  def chmod(self, path: str, mode: int):
    """Change the file permissions.

    Args:
      path: The path to the file.
      mode: The new file mode permissions
    """
    logger.info('chmod %s %s', path, mode)
    file = self.root.file_at_path(path)
    file.chmod(mode)

  def chown(self, path: str, uid: str, gid: str):
    """Change the file owner.

    Args:
      path: The path to the file.
      uid: The user owner id.
      gid: The group owner id.
    """
    logger.info('chown %s %s %s', path, uid, gid)
    file = self.root.file_at_path(path)
    file.chown(uid, gid)

  def create(self, path: str, mode: int) -> int:
    """Create an empty file.

    Args:
      path: The path to the file to create.
      mode: The permissions on the file.

    Returns:
      The file descriptor.
    """
    logger.info('create %s %s', path, mode)
    file = self.root.touch(path, mode)
    self.cache.add(file.file_id, self._to_bytes(''))
    return self.open()

  def open(self, _=None, __=None) -> int:
    """Increment the file descriptor.

    Returns:
      A new file descriptor.
    """
    self.fd += 1
    return self.fd

  def getattr(self, path: str, _) -> Dict:
    """
    Args:
      path: The path to the file.

    Returns:
      The file metadata.
    """
    if not self.root.file_exists(path):
      raise FuseOSError(ENOENT)
    return self.root.file_at_path(path).metadata

  def getxattr(self, path: str, name: str, _) -> str:
    """Read a file attribute.

    Args:
      path: The path to the file.
      name: The name of the attribute to read.
    
    Returns:
      The value of the attribute for the file.
    """
    file = self.root.file_at_path(path)
    if name in file.attrs:
      return file.attrs[name]
    return ''.encode('utf-8')

  def listxattr(self, path: str) -> List[str]:
    """
    Args:
      path: The path to the file.

    Returns:
      The file's extra attributes.
    """
    file = self.root.file_at_path(path)
    return file.attrs.keys()

  def mkdir(self, path: str, mode: int):
    """Create a new directory.

    Args:
      path: The path to create.
      mode: The directory permissions.
    """
    logger.info('mkdir %s %s', path, mode)
    self.root.mkdir(path, mode)

  def read(self, path: str, size: int, offset: int, _=None) -> str:
    """Read the file's contents.

    Args:
      path: Theh path to the file to read.
      size: The number of bytes to read.
      offset: The offset to read from.

    Returns:
      The queried bytes of the file.
    """
    logger.info('read %s %s %s', path, offset, size)
    file = self.root.file_at_path(path)
    logger.info('Found file %s', file.file_id)

    if file.st_size == 0:
      # Special case for empty files
      logger.info('File size %s', 0)
      return self._to_bytes('')

    with self.file_locks[file.file_id]:
      # Download from the object store if the file is not cached
      if not self.cache.has(file.file_id):
        logger.info('File not in cache, downloading from store')
        contents = self._to_bytes(self.b2.download_file(file.file_id))
        logger.info('File downloaded %s', len(contents))
        self.cache.add(file.file_id, contents)

      content = self.cache.get(file.file_id)
      logger.info('File size %s', len(content))
      return content[offset:offset + size if size else None]

  def readdir(self, path: str, _) -> List[str]:
    """Read the entries in the directory.

    Args:
      path: The path to the directory.
    
    Returns:
      The names of the entries (files and subdirectories).
    """
    logger.info('readdir %s', path)
    dir = self.root.file_at_path(path)
    return ['.', '..'] + [f for f in dir.files]

  def readlink(self, path: str) -> str:
    """Read the entire contents of the file.

    Args:
      path: The file to read.

    Returns:
      The file's contents.
    """
    logger.info('readlink %s', path)
    return self.read(path, None, 0)

  def removexattr(self, path: str, name: str):
    """Remove an attribute from a file.

    Args:
      path: Path to the file.
      name: Name of the attribute to remove.
    """
    file = self.root.file_at_path(path)
    if name in file.attrs:
      del file.attrs[name]

  def rename(self, old: str, new: str):
    """Rename a file by deleting and recreating it.

    Args:
      old: The old path of the file.
      new: The new path of the file.
    """
    logger.info('rename %s %s', old, new)
    file = self.root.file_at_path(old)
    if type(file) == Directory:
      if len(file.files) > 0:
        logger.info('Directory not empty')
        return ENOTEMPTY
      self.rmdir(old)
      self.mkdir(new, file.st_mode)
    else:
      contents = self.readlink(old)
      self.unlink(old)
      self.create(new, file.st_mode)
      self.write(new, contents, 0)

  def rmdir(self, path):
    """Remove a directory, if it is not empty.

    Args:
      path: The path to the directory.
    """
    logger.info('rmdir %s', path)
    directory = self.root.file_at_path(path)
    if len(directory.files) > 0:
      return ENOTEMPTY
    self.root.rm(path)

  def setxattr(self, path: str, name: str, value: str, _, __):
    """Set an attribute for the file.

    Args:
      path: Path to the file.
      name: Name of the attribute to set.
      value: Value of the attribute.
    """
    file = self.root.file_at_path(path)
    file.attrs[name] = value

  def statfs(self, _):
    """Get file system stats."""
    return dict(f_bsize=4096, f_blocks=4294967296, f_bavail=4294967296)

  def symlink(self, target, source):
    """Symlink from a target to a source.

    Args:
      target: The symlinked file.
      source: The original file.
    """
    # No support for symlinking
    return FuseOSError(EINVAL)

  def truncate(self, path: str, length: int, _):
    """Truncate or pad the file to the specified length.

    Args:
      path: The file to truncate.
      length: The desired lenght.
    """
    file = self.root.file_at_path(path)
    content = self.readlink(path)
    content = content.ljust(length, '\x00'.encode('utf-8'))
    file.st_size = length

  def _delete_file(self, path: str):
    """Delete a file from both the local cache and the object store.

    Args:
      path: The path to the file.
    """
    file = self.root.file_at_path(path)
    if self.cache.has(file.file_id):
      self.cache.delete(file.file_id)
    if file.st_size > 0:
      self.b2.delete_file(file.file_id, path.strip('/'))

  def unlink(self, path: str):
    """Delete a file.

    Args:
      path: The path to the file.
    """
    logger.info('unlink %s', path)
    file = self.root.file_at_path(path)
    if type(file) == Directory:
      self.rmdir(path)
    else:
      self._delete_file(path)
      self.root.rm(path)

  def utimens(self, path: str, times: Tuple[int, int] = None):
    """Update the touch time for the file.

    Args:
      path: The file to update.
      times: The modify times to apply.
    """
    file = self.root.file_at_path(path)
    now = time()
    mtime, atime = times if times else (now, now)
    file.update(modify_time=mtime, access_time=atime)

  def _upload_file(self, path: str) -> str:
    """Upload a file to the object store.

    Args:
      path: The path of the file to upload.

    """
    logger.info('upload %s', path)
    file = self.root.file_at_path(path)
    content = self.cache.get(file.file_id)

    with self.file_locks[file.file_id]:
      logger.info('Uploading file %s', len(content))
      response = self.b2.upload_file(self.bucket_id, path.strip('/'), content)
      logger.info('Upload complete')

      logger.info('Updating cache')
      self.cache.delete(file.file_id)
      file.update(file_id=response['fileId'], file_size=len(content))
      self.cache.add(file.file_id, content)

  def write(self, path: str, data: str, offset: str, _=None) -> int:
    """Write data to a file.

    Args:
      path: The file to write to.
      data: The bytes to write.
      offset: The offset in the file to begin writing at.
    
    Returns:
      The number of bytes written.
    """
    logger.info('write %s %s %s', path, offset, len(data))
    file = self.root.file_at_path(path)
    content = self.readlink(path)

    with self.file_locks[file.file_id]:
      # Delete the exisiting version of the file if it exists
      # self._delete_file(path)

      # Write the new bytes
      data = self._to_bytes(data)
      content = (content[:offset].ljust(offset, self._to_bytes('\x00')) + data +
                 content[offset + len(data):])

      # Immediately save locally
      logger.info('Saving to cache %s %s', file.file_id, len(content))
      file.update(file_size=len(content))
      self.cache.add(file.file_id, content)

      # Submit task to upload to object store
      self.task_queue.submit_task(file.file_id, self.upload_delay,
                                  self._upload_file, path)

      return len(data)



from setuptools import setup


def read(file_name):
  with open(file_name) as f:
    return f.read()


setup(name='zerofs',
      version='0.1.5',
      description='Transparant filesystem backed by Backblaze B2 object store',
      long_description=read('README.md'),
      url='https://github.com/picklelo/zerofs',
      author='Nikhil Rao',
      author_email='nikhil@nikhilrao.me',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3'
      ],
      keywords='backblaze b2 zero fs filesystem',
      packages=['zerofs'],
      scripts=['bin/zerofs'],
      install_requires=['backblazeb2', 'fusepy'])



from abc import ABC, abstractmethod
from collections import defaultdict
from stat import S_IFDIR, S_IFREG
from time import time
from typing import Dict, List, Union
from uuid import uuid4


class FileBase(ABC):
  """Abstract base class for file-like objects."""

  def __init__(self, name: str):
    self.name = name
    self.st_mode = None
    self.st_uid = None
    self.st_gid = None
    self.attrs = {}

  def chmod(self, mode):
    self.st_mode &= 0o770000
    self.st_mode |= mode

  def chown(self, uid, gid):
    self.st_uid = uid
    self.st_gid = gid


class File(FileBase):
  """Represents a file backed by the object store."""

  def __init__(self, file: Dict):
    """Create a file object.

    Args:
      file: A dictionary of file metadata from B2.
    """
    super().__init__(file.get('fileName', ''))
    self.file_id = file.get('fileId', '')
    self.st_size = file.get('contentLength', 0)
    self.st_mtime = file.get('uploadTimestamp', time() * 1e3) * 1e-3
    self.st_ctime = self.st_mtime
    self.st_atime = self.st_mtime
    self.st_mode = S_IFREG | 0o755
    self.st_nlink = 1

  def __repr__(self):
    return '<File {}>'.format(self.name)

  @property
  def metadata(self) -> Dict:
    return {
        'st_mode': self.st_mode,
        'st_ctime': self.st_ctime,
        'st_mtime': self.st_mtime,
        'st_atime': self.st_atime,
        'st_nlink': self.st_nlink,
        'st_size': self.st_size
    }

  def update(self,
             file_id: str = None,
             file_size: int = None,
             modify_time: str = None,
             access_time: int = None):
    """Update the file metadata.
    Automatically updates the last modified time.

    Args:
      file_id: The new file id.
      file_size: The new file size.
    """
    if file_id:
      self.file_id = file_id
    if file_size:
      self.st_size = file_size
    if modify_time:
      self.st_mtime = modify_time
    if access_time:
      self.st_atime = access_time


class Directory(FileBase):
  """A virtual directory containing subfiles and directories."""

  def __init__(self, name: str, files: List[File], mode=0o755):
    """Initialize with a list of files in this directory.

    Args:
      name: The name of the directory.
      files: A list with file metadata for files in the directory.
      mode: The permissions to set.
    """
    super().__init__(name)

    children = defaultdict(list)
    for file in files:
      filename = file.name.strip('/')
      parts = filename.split('/', 1)
      if len(parts) > 1:
        # This file is in a subdirectory
        parent, path = parts
      else:
        # This file is in this directory
        parent, path = '', filename
      file.name = path
      children[parent].append(file)

    self.files = {file.name: file for file in children['']}
    del children['']
    self.files.update({k: Directory(k, v) for k, v in children.items()})

    self.st_mode = S_IFDIR | mode
    self.st_atime = time()

  @property
  def st_mtime(self) -> float:
    """The last modified time of the directory."""
    if len(self.files) == 0:
      return self.st_atime
    return max([f.st_mtime for f in self.files.values()])

  @property
  def st_nlink(self) -> int:
    """Number of hard links pointing to the directory."""
    return 2 + len([f for f in self.files if type(f) == Directory])

  @property
  def metadata(self) -> Dict:
    return {
        'st_mode': self.st_mode,
        'st_ctime': self.st_mtime,
        'st_mtime': self.st_mtime,
        'st_atime': self.st_mtime,
        'st_nlink': self.st_nlink
    }

  @staticmethod
  def _to_path_list(path: Union[str, List[str]]) -> List[str]:
    """Combine a path to a path list.

    Args:
      path: The path to convert (can be a string or a list)

    Returns:
      A list that can be used to find the node in the tree.
    """
    if type(path) == str:
      path = path.strip('/').split('/')
    return path

  def file_nesting(self, path: Union[str, List[str]]) -> List['Directory']:
    """Get the nesting directories of a file or directory.

    Args:
      path: The path to get the nesting in.

    Returns:
      A list of directories indicating the nested structure of the path.
    """
    path = self._to_path_list(path)
    if path[0] == '':
      return [self]
    file = self.files[path[0]]
    if len(path) == 1:
      # The file is in the root directory
      return [self, file]
    return [self] + file.file_nesting(path[1:])

  def file_at_path(self, path: Union[str, List[str]]) -> File:
    """Get the file given a path relative to this directory.

    Args:
      path: The path of the file to query.

    Returns:
      The file object at the path if it exists.
    """
    return self.file_nesting(path)[-1]

  def file_exists(self, path: str) -> bool:
    """
    Args:
      path: The path to the file to check.

    Returns:
      Whether the file exists.
    """
    try:
      self.file_at_path(path)
      return True
    except KeyError:
      return False

  def _find_node(self, path: Union[str, List[str]]) -> FileBase:
    """Find the node in the directory tree.

    Args:
      path: The path to the node to find.

    Returns:
      The found node.
    """
    path = self._to_path_list(path)
    if len(path) == 0:
      return self

    if (path[0] not in self.files or type(self.files[path[0]]) != Directory):
      raise KeyError('Cannot find node, directory {} does not exist'.format(
          path[0]))

    return self.files[path[0]]._find_node(path[1:])

  def mkdir(self, path: Union[str, List[str]], mode: int):
    """Create a subdirectory.

    Args:
      path: The path to the directory to create.
      mode: The directory permissions.
    """
    path = self._to_path_list(path)
    node = self._find_node(path[:-1])
    if path[-1] in node.files:
      raise KeyError('Directory {} already exists'.format(path))
    node.files[path[-1]] = Directory(path[-1], [], mode=mode)

  def rm(self, path: Union[str, List[str]]):
    """Remove a file or directory.

    Args:
      name: The path to the file.
    """
    nesting = self.file_nesting(path)
    if len(nesting) < 2:
      raise ValueError('Cannot rm the root directory')
    parent_dir, file = nesting[-2:]
    del parent_dir.files[file.name]

  def touch(self, path: Union[str, List[str]], mode: int) -> File:
    """Create an empty file.

    Args:
      path: The path to the file to create.
      mode: The file permissions.
    """
    path = self._to_path_list(path)
    node = self._find_node(path[:-1])
    file = File({'fileId': str(uuid4()), 'fileName': path[-1]})
    node.files[path[-1]] = file
    return file
